/*-----------------------------------------------------------------------------
Modul Name : BaseAgent.cpp
Date       : 30.10.97
-------------------------------------------------------------------------------
Author     : Klaus Dorer

Description: contains the operations for a basic agent

Comments   : _
-----------------------------------------------------------------------------*/

#include "OS.h"
#include "stdafx.h"
#include "Player.h"
#include "Agentsock.h"
#include "BaseAgent.h"
#include "PlayerDoc.h"


#include <math.h>


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


#define MAX_MSG_LENGTH 1500

/*-----------------------------------------------------------------------------
Function: CBaseAgent::CBaseAgent()
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: none
Description: constructor
-----------------------------------------------------------------------------*/
CBaseAgent::CBaseAgent()
{
	m_iPositionX = 0;
	m_iPositionY = -30;
	m_iDirection = 0;
	m_bConnected = FALSE;
	m_iSide = LEFT_TO_RIGHT;
	m_iNumber = 0;

	m_iAutoPlay = 0;
}

CBaseAgent::~CBaseAgent()
{
	DisconnectAgent();
}

/*-----------------------------------------------------------------------------
Function: BOOL CBaseAgent::ConnectAgent(LPCTSTR lpszTeam, LPCTSTR lpszServer, UINT nPort)
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: name of the team
			  name of the server (URL)
			  port for the socket
Returnval: TRUE, if the agent could be created and connected to a server socket
			  FALSE otherwise
Description: creates the socket of the agent and initializes it
-----------------------------------------------------------------------------*/
BOOL CBaseAgent::ConnectAgent(LPCTSTR lpszTeam, LPCTSTR lpszServer, UINT nPort,BOOL bGoalie)
{
	// is the agent already connected (should not happen) 
	if (m_bConnected)
		return TRUE;

	// create the datagram socket
	if (!Create(nPort, lpszServer))
	{
		// socket could not be created
		OnErrorCreatingSocket();
		return FALSE;
	}
	else
	{
		m_bConnected = TRUE;
		m_strTeam = lpszTeam;
		m_bGoalie = bGoalie;
		OnCreatedSocket();
		return TRUE;
	}
}

/*-----------------------------------------------------------------------------
Function: void OnCreatedSocket()
Date    : 17.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: notification handler of a succesful socket creation
-----------------------------------------------------------------------------*/
void CBaseAgent::OnCreatedSocket()
{
	CString	message;

	// send the (init team) command
	if (m_bGoalie) message.Format("(init %s (goalie) (version 4.00))", (LPCTSTR) m_strTeam);
	else message.Format("(init %s (version 4.00))", (LPCTSTR) m_strTeam);
	
	SendMsg(message);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::DisconnectAgent()
Date    : 16.12.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: empties the agent's socket and deletes it
-----------------------------------------------------------------------------*/
void CBaseAgent::DisconnectAgent()
{
	// is the socket active ?
	if (m_bConnected)
	{
		SendBye();

		BYTE Buffer[50];

		ShutDown();
		while(Receive(Buffer,50) > 0);
		Close();
	}
	m_bConnected = FALSE;
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::SendMsg(CString& strText)
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: message to be sent
Returnval: (void)
Description: the message is sent to the socket by SendTo (unconnected socket)
-----------------------------------------------------------------------------*/
void CBaseAgent::SendMsg(CString& strText)
{
	int		iLength;
	
	// send the buffer
	iLength = SendTo(strText, strText.GetLength(), GetPort(), GetAddress(), 0);
	if (iLength == strText.GetLength())
		OnSentMessage((LPCTSTR) strText);
	else
	{
		// an error occured
		#ifdef _DEBUG
			afxDump << "error writing to socket\n";
		#endif
		OnErrorSendingMessage();
	}
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::ReceiveMsg()
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: a message is received from the socket and processed by the agent
-----------------------------------------------------------------------------*/
void CBaseAgent::OnReceive(int iError)
{
	char		buffer[MAX_MSG_LENGTH];
	int		iLength = 0;
	UINT		iPort = GetPort();
	int		iPosition = 0;

	// first call base classes method
	CAgentSocket::OnReceive(iError);

	// read a message from the socket
	iLength = ReceiveFrom(buffer, MAX_MSG_LENGTH, GetAddress(), iPort, 0);
	// the port may have changed
	SetPort(iPort);

	if (iLength > 0)
		OnReceivedMessage((LPCTSTR) buffer);
	else
	{
		// an error occured
		#ifdef _DEBUG
			afxDump << "error reading socket\n";
		#endif
		OnErrorReceivingMessage();
	}
}

void CBaseAgent::OnReceivedMessage(LPCTSTR message)
{
        m_pPerception.m_strTeam = m_strTeam;
		m_pPerception.ParseMessage(message);
		// obnovenie pohadu
		switch(message[1])
		{
		case 'h':
			m_pDoc->OnHear(message);
			break;
		case 's':
			// something to see or sense body
			if (message[3] == 'e')
				m_pDoc->OnSee();
			else
				// a sense body information
				m_pDoc->OnSenseBody();
			break;
		}
		POSITION pos = m_pDoc->GetFirstViewPosition();
		CView* pFirstView = m_pDoc->GetNextView( pos );
		pFirstView->Invalidate(FALSE);

}


/*-----------------------------------------------------------------------------
Function: void CBaseAgent::Move(int x, int y)
Date    : 31.10.97
-------------------------------------------------------------------------------
Parameter: x,y coordinate (in meters) where to move to before playing
Returnval: (void)
Description: following functions implement the abilities of soccer agents
				 see the soccer manual for more details
-----------------------------------------------------------------------------*/
void CBaseAgent::Move(int x, int y)
{
	CString	message;

	message.Format("(move %d %d)", x, y);
	SendMsg(message);

	m_iPositionX = x;
	m_iPositionY = y;
}

void CBaseAgent::Turn(int angle)
{
	CString	message;

	// angle may be -180 to 180
	message.Format("(turn %d)", angle);
	SendMsg(message);

	m_iDirection += angle;
	if (m_iDirection > 180)
		m_iDirection -= 360;
}

void CBaseAgent::Dash(int power)
{
	CString	message;

	// power may be -30 to 100
	message.Format("(dash %d)", power);
	SendMsg(message);
}

void CBaseAgent::Kick(int power, int direction)
{
	CString	message;
	
	// power -30 to 100
	// direction -180 to 180
	message.Format("(kick %d %d)", power, direction);
	SendMsg(message);
}

void CBaseAgent::Catch(int direction)
{
	CString	message;
	
	// direction -180 to 180
	message.Format("(catch %d)", direction);
	SendMsg(message);
}

void CBaseAgent::Say(CString msg)
{
	CString	message;

	// max 255 byte
	message.Format("(say ");
	message += msg;
	message += _T(")");
	SendMsg(message);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::ChangeView(int width, int quality)
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: view angle of the agent (45, 90 or 180 degrees)
			  view quality (high, low)
Returnval: (void)
Description: changes the agents view mode. With 45 deg the agent gets perceptions
				 from the server every 75ms, with 90 every 150ms and with 180 every 300ms.
-----------------------------------------------------------------------------*/
void CBaseAgent::ChangeView(int width, int quality)
{
	CString	message;

	message =_T("(change_view ");
	// set the view angle string
	if (width == 180)
		message += _T("wide ");
	else if (width == 90)
		message += _T("normal ");
	else
		message += _T("narrow ");
	// set the view quality string
	if (quality == 0)
		message += _T("low)");
	else
		message += _T("high)");

	SendMsg(message);
	OnChangedView(width, quality);
}

void CBaseAgent::SenseBody()
{
	CString	msg("(sense_body)");

	SendMsg(msg);
}

/*-----------------------------------------------------------------------------
Function: void CBaseAgent::SendBye()
Date    : 08.03.99
-------------------------------------------------------------------------------
Parameter: none
Returnval: (void)
Description: disconnects the client from the server
-----------------------------------------------------------------------------*/
void CBaseAgent::SendBye()
{
	CString	msg("(bye)");

	SendMsg(msg);
}

void CBaseAgent::MoveRel(int amount)
{
	int	x;
	int	y;

	x = m_iPositionX + (int) cos(m_iDirection * PI / 180);
	y = m_iPositionY + (int) sin(m_iDirection * PI / 180);
	// and move to the calculated position
	Move (x, y);		
}

// Rado
TVector CBaseAgent::GetBallInterceptPos(TPlayer* pAgent)
{
	int				i;
	int				delay;
	int				power = 100;
	double			dActionRadius;
	double			dBallDistance;
	TVector			tFutureBall;
	TVector			tPrevious;
	TDirection angle; 
//	TPlayer*		pThisPlayer = m_pPerception.GetThisPlayer();
	BOOL			bReady = FALSE;

	// calculate direction where to run to
	// delay are the number of cycles that may get lost by turns and other things
	// the bigger delay, the more cautious are the expectations of the agent
	delay = 1;  // 3  -> nam staci jeden turn
	i = delay;

	dActionRadius = pAgent->GetFutureDistanceActive(0, power, pAgent->GetAbsDir(), 
													m_pPerception.GetEffort());
	tFutureBall = m_pPerception.GetBall()->GetPosition();
	TDirection angle0 = pAgent->GetAngleTo(tFutureBall);

	while (!bReady && (i < MAX_FUTURE_POSITIONS))
	{
		// we calculate the intersection of the future agent position and ball position
		tPrevious = tFutureBall;
		tFutureBall = *(m_pPerception.GetFutureBallPosition(i));
		dBallDistance = pAgent->GetDistanceTo(tFutureBall);

		// nemozeme utekat do miesta, smerom ku ktoreme nevidiet loptu
		angle = angle0 - pAgent->GetAngleTo(tFutureBall);
		if ((angle > 40) || (angle < -40)) {
			tFutureBall = tPrevious;  break;
		}

/*		// test if the ball will be outside
		if (m_pPerception->IsOutside(tFutureBall))
		{
			// we run to the last position that is within the field
			// !!! may be if the two distances are big, do not run anyhow
			tFutureBall = tPrevious;
			bReady = TRUE;
		}
		else  */
		{
			// get distance the agent can dash
			dActionRadius = pAgent->GetNextFutureDistance();

			// pridane + 0.7 = kickable_margin  // XXX
			if (dActionRadius + 0.7 > dBallDistance)
				// we have the position
				bReady = TRUE;
			else
			{
				// ball will still be too far this setp
//				tPrevious = tFutureBall;
				i++;
			}
		}
	}

	return tFutureBall;
}

int CBaseAgent::GetAvoidAngle(TPlayer* pAgent, TVector DestPos)
{
	int i;
	int minangle, maxangle, angle, angle1, angle2, dangle;
	double dist;
	TVector pos;

	TVector pos0 = pAgent->GetPosition();
	int angle0 = pAgent->GetAngleTo(DestPos);
	double dist0 = (DestPos - pos0).GetLength();

	minangle = maxangle = 0;
	for (i = 0; i < m_pPerception.GetNumOfRegPlayers(); i++) {
		pos = m_pPerception.GetRegPlayers()[i]->GetPosition();
		dist = (pos - pos0).GetLength();
		if (dist < dist0) {
			angle = pAgent->GetAngleTo(pos) - angle0;
//			if (dist < 0.7) 
//				dangle = 40; else // 90
			if (dist < 2) dangle = 30;  // 1.2) dangle = 30;
			else if (dist < 8) dangle = 10;
			else dangle = 0;
			angle1 = angle - dangle;
			angle2 = angle + dangle;
			if (((angle1 > 0) && (angle2 > 0)) || ((angle1 < 0) && (angle2 < 0))) angle1 = angle2 = 0;

			if (maxangle < angle2) maxangle = angle2;
			if (minangle > angle1) minangle = angle1;
		}
	}

	return ((-minangle < maxangle)?(minangle):(maxangle));
}

int CBaseAgent::RunTo(TVector DestPos, int Backw)
{
	TPlayer* pThisPlayer = m_pPerception.GetThisPlayer();
	int minDir;
	double dist = pThisPlayer->GetDistanceTo(DestPos);

	if (dist < 1) return 1;					// uz som dobehol

	if (dist > 5) minDir = 8;
//	else  minDir = (int) (asin(1.0 / (dist + 1.0)) * RAD2DEG);  //???
	else  minDir = (int) (0.5 + asin(0.7 / dist) * RAD2DEG);

	int dir = pThisPlayer->GetAngleTo(DestPos);	
	int avoiddir = GetAvoidAngle(pThisPlayer, DestPos);
	if (avoiddir != 0) minDir = 8;
	dir += avoiddir;

	if (Backw) {
		// je to pripustny uhol pre beh dozadu??
		if ((dir < -150) || (dir > 150)) {
			dir += 180;
			if ((dir < -minDir) || (dir > minDir)) 
				Turn(dir);
			else
				Dash(-30);
			return 0;
		}
	} 

	if ((dir < -minDir) || (dir > minDir)) 
		Turn(dir);
	else
		Dash(100);

	return 0;
}

/*-----------------------------------------------------------------------------
Function: void TDoingAgent::ComplexKick(int iPower, int iDirection)
Date    : 15.07.98
-------------------------------------------------------------------------------
Parameter: power and relative direction in which to kick
Returnval: (void)
Description: avoids to kick onto the agent itself and takes the speed of the 
				 ball into account
-----------------------------------------------------------------------------*/
int CBaseAgent::KickRel0(int Power0, TDirection Dir0)
{
	TBall*			pBall = m_pPerception.GetBall();
	TPlayer*	    pAgent = m_pPerception.GetThisPlayer();
	double			dBallDist = pBall->GetDistance() - PLAYER_SIZE - BALL_SIZE;
	double			dDeltaDir = (double) pBall->GetDirection() / 180.0;
    TVector			tBallSpeed = pBall->GetRelSpeed(pAgent->GetPosition(), pAgent->GetAbsDir());
	// kedze pocitame uhly zrkadlovo, tak prevrat
	tBallSpeed.SetY(-tBallSpeed.GetY());
	double			dBallSpeed = tBallSpeed.GetLength();
	TVector			tWantToKick;
	double			dFactor;
//	TVector         tPlayerSpeed = pAgent->GetAbsSpeed();

	int				Power;
	TDirection		Dir;

	// ak sa hrac pohybuje prilis rychlo, tak nevieme spravne urcit rychlost lopty 
	if (pAgent->GetAbsSpeed().GetLength() > 0.01) return -1;
	
	if (dDeltaDir < 0)
		dDeltaDir = -dDeltaDir;

	// calculate the wanted final ball speed
	dFactor = KICK_POWER_RATE * (1.0 - 0.25 * dDeltaDir - 0.25 * dBallDist / KICKABLE_MARGIN);
	tWantToKick.SetVectorByDirection(Power0 * dFactor, (int) Dir0);

	if (Power0 >= MAXPOWER)
	{
		// a full power kick does not slow down the ball if it already goes into correct dir
		// so we assume that the want to kick is ballspeed + 100 power
		tWantToKick *= ((dBallSpeed + MAXPOWER * KICK_POWER_RATE) / tWantToKick.GetLength());
		tWantToKick -= tBallSpeed;
	}
	else
		tWantToKick -= tBallSpeed; 

	Dir = tWantToKick.GetDirection();
	Power = (int) (tWantToKick.GetLength() / dFactor);

	if (Power > MAXPOWER)
		Power = int(MAXPOWER + 0.5);
	if (Power <= 0) 
		return 1;

	Kick(Power, Dir);
	
	return 0;
}

int CBaseAgent::KickRel(int Power0, TDirection Dir0)
{
	// ak nevieme kopnut relativne, kopneme absolutne
	if (KickRel0(Power0, Dir0) < 0) 
		Kick(Power0, Dir0);
	return 0;
}

int CBaseAgent::KickTo(TDirection DestDir)
{
	TDirection dir = DestDir - (m_pPerception.GetBallDirection() + 180);

//	if ((dir >= 0) && (dir < 45))
//		Kick(20,  -90);
//	else if ((dir <= 0) && (dir > -45)) 
//		Kick(20, 90); 
//	else 
//		Kick(100, DestDir);

	if (!((dir < -45) || (dir > 45)))
		TurnBallTo(DestDir, 0);
	else 
		KickRel(100, DestDir);

	return 0;
}

int CBaseAgent::TurnBallTo(TDirection DestDir, int Safe)
{
	static int KickCount = 0;

	int power;
	double sign, val;
	TPlayer* pThisPlayer = m_pPerception.GetThisPlayer();

	double dist = m_pPerception.GetBallDistance();
	TDirection dir = m_pPerception.GetBallDirection();
	int ddir = int(DestDir - dir);

	// je lopta na svojom mieste s dostatocnou presnostou??
	if (!((ddir < -5) || (ddir > 5))) return 1;

	sign = (ddir > 0)?(1.0):(-1.0);

	if (dist > 0.6)	val = 110.0;
	else if (dist < 0.4) val = 95.0;
	else val = 90.0;

//	TVector ballspeed = m_pPerception.GetBall()->GetAbsSpeed();
//	if (ballspeed.GetLength() > 0.2) {
//		// stopni loptu
//		TDirection ang1 = ballspeed.GetDirection() + 180;
//		TDirection ang2 = pThisPlayer->GetAbsDir();
//		Kick(1, (int)double(ang2 - ang1));
//		Say("yyy");
//	} else

	if (ddir < 0) ddir = -ddir;
	if (ddir < 10) power = 1;
	else if (ddir < 30) power = 6;
	else power = 11;

	if (Safe) {
		if (KickRel0(power, dir + TDirection(int(sign * val))) < 0)
			return -1;
	} else {
		return KickRel(power, dir + TDirection(int(sign * val)));
	}

	return 0;
}

TPlayer* CBaseAgent::GetNearestOpponent(TPlayer* pAgent0)
{
	int i;
	double mindist, dist;
	TPlayer *pAgent, *pMinAgent;

	TVector pos0 = pAgent0->GetPosition();

	pMinAgent = NULL;  mindist = 10000.0;
	for (i = 0; i < m_pPerception.GetNumOfRegPlayers(); i++) {
		pAgent = m_pPerception.GetRegPlayers()[i];
		if (!pAgent->IsOwnTeam()) {
			dist = (pAgent->GetPosition() - pos0).GetLength();
			if (mindist > dist) {
				mindist = dist;
				pMinAgent = pAgent;
			}
		}
	}

	return pMinAgent;
}

TVector CBaseAgent::GetDefendPos(TPlayer* pAgent)
{
	TVector goalpos = m_pPerception.GetOurGoal()->GetPosition();

	return (pAgent->GetPosition() - goalpos)*0.66 + goalpos;
}

//double FormX1[11] = { -4*FW/5, -FW+PW, -FW+PW, -FW+PW,     0,      0,    0,      0, FW/2,  FW/2,  FW/2};
//double FormY1[11] = {       0,      0,   FH/3,  -FH/3,  FH/4,  -FH/4, FH/2,  -FH/2,    0,  FH/3, -FH/3};

double FormX1[11] = { -4*FW/5,     0,      0, FW/2, -FW+PW, FW/2,  FW/2,    0,      0, -FW+PW, -FW+PW};
double FormY1[11] = {       0,  FH/4,  -FH/4,    0,      0, FH/3, -FH/3, FH/2,  -FH/2,   FH/3,  -FH/3};

double FormX2[11] = { -4*FW/5,     0,      0, FW/2, -FW+PW, FW/2,  FW/2,    0,      0, -FW+PW, -FW+PW};
double FormY2[11] = {       0,  FH/4,  -FH/4,    0,      0, FH/3, -FH/3, FH/2,  -FH/2,   FH/3,  -FH/3};

int CBaseAgent::GetFormationPos(TVector &defpos) 
{
	int idx = m_pPerception.GetNumber() - 1;
	double a1, a2, c;

	a1 = m_pPerception.GetBall()->GetPosition().GetX();
	c = (a1 + FW)/(2*FW);  if (c > 1) c = 1;  if (c < 0) c = 0; 

	a1 = FormX1[idx]; a2 = FormX2[idx];
	defpos.SetX(a1 + (a2 - a1) * c);
	a1 = FormY1[idx]; a2 = FormY2[idx];
	defpos.SetY(a1 + (a2 - a1) * c);

	if (m_pPerception.GetSide() != LEFT_TO_RIGHT) defpos.SetX(-defpos.GetX());

	return 0;
}


#define SayInt(i) {	char buffer[100]; sprintf(buffer, "%d", i); Say(buffer); }

void CBaseAgent::AutoPlay1()
{
	static int cont = 0;
	static int letsee = 0;
	static int cnt = 0;
	static TVector contrun;
	TPlayer* pAgent;
	TPlayer* pThisPlayer = m_pPerception.GetThisPlayer();

//m_aPlayer.Turn(30);
//return;

	if (!m_pPerception.GetSeeBall()) {
		// ak nevidim loptu, ale viem, kam mam utekat, tak utekam
		if (cont > 0) {
			cont--;
			if (RunTo(contrun) == 0) return;
			cont = 0;
		}


		if (letsee) Turn(pThisPlayer->GetAngleTo(m_pPerception.GetBall()));
		else Turn(-45);
		letsee = 0;		
	} else {
		letsee = 1;
		double dist = pThisPlayer->GetDistanceTo(m_pPerception.GetBall());
		TVector ballspeed = m_pPerception.GetBall()->GetAbsSpeed();

		int i, minidx, minidx2;
		double mindist, mindist2, bidist, bidist2;
		TVector mybipos = GetBallInterceptPos(pThisPlayer);

		minidx = -1; minidx2 = -1;
		TVector newpos = mybipos;
		TVector minpos;
		mindist2 = mindist = (pThisPlayer->GetPosition() - newpos).GetLength();

		m_pPerception.mm_x2 = (float)newpos.GetX();
		m_pPerception.mm_y2 = (float)newpos.GetY();
		
		// zisti, ktory hrac nasho timu je najblizsie k chyteniu lopty
		for (i = 0; i < m_pPerception.GetNumOfRegPlayers(); i++) {
			pAgent = m_pPerception.GetRegPlayers()[i];
			if (pAgent->IsOwnTeam()) {
				newpos = GetBallInterceptPos(pAgent);
				bidist = (pAgent->GetPosition() - newpos).GetLength();
				if (mindist > bidist + 1.0) { mindist = bidist; minidx = i;	minpos = newpos; }
				// SayInt(i);  SayInt(int(100*pThisPlayer->GetAbsSpeed().GetLength()));
			} else {
				newpos = GetBallInterceptPos(pAgent);
				bidist2 = (pAgent->GetPosition() - newpos).GetLength();
				if (mindist2 > bidist2 + 1.0) { mindist2 = bidist2; minidx2 = i; }
			}
		}

		if ((minidx >= 0) && (minidx2 >= 0)) {
			// ak je super blizsie k lopte 
			if (mindist2 < mindist) {
				// a nas najblizsi je z druhej strany
				if ((m_pPerception.GetRegPlayers()[minidx]->GetPosition() - pThisPlayer->GetPosition()).GetLength() >
					(m_pPerception.GetBall()->GetPosition() - pThisPlayer->GetPosition()).GetLength()) {
					minidx = -1;
				}
			}
		}

		if (minidx == -1) {
			// ja som zodpovedny za loptu
			if ((dist < 1.0) || (RunTo(mybipos))) {
				// toc loptu, len ked stojime
				//			if (pThisPlayer->GetAbsSpeed().GetLength() < 0.01) 
				//			if (cnt++ > 8) { Turn(90); cnt = 0; }
				//			else
				////ChangeView(180, 1);
				////if ((pAgent = GetNearestOpponent(pThisPlayer)) != NULL) 
				////	newpos = pAgent->GetPosition();
				////else
				////	newpos = m_pPerception.GetOurGoal()->GetPosition();
				////if (TurnBallTo(pThisPlayer->GetAngleTo(newpos) + 180) > 0)
				////	Turn(pThisPlayer->GetAngleTo(newpos));
				KickTo(m_pPerception.GetGoalDirection());
			} else cnt = 0;
		} else {
			// o loptu sa nemusim starat
			TVector defpos;

			minidx = -1;
/*
			// zisti, ktory hrac super. timu je "volny"
			//   s najmensou vzd. k chyteniu lopty + ja k jeho obrane
			for (i = 0; i < m_pPerception.GetNumOfRegPlayers(); i++) {
				pAgent = m_pPerception.GetRegPlayers()[i];
				// xxx + test, ci je volny
				if (pAgent->IsValid() && !pAgent->IsOwnTeam()) {
					newpos = GetBallInterceptPos(pAgent);
					bidist = (pAgent->GetPosition() - newpos).GetLength();
					newpos = GetDefendPos(pAgent);
					bidist += (pThisPlayer->GetPosition() - newpos).GetLength();
					if ((minidx < 0) || (mindist > bidist)) { mindist = bidist; minidx = i; defpos = newpos; }
					// SayInt(i);  SayInt(int(100*pThisPlayer->GetAbsSpeed().GetLength()));
				}
			}
*/
			// miesto obrannej pozicie sa postavime 
			//   na nase miesto vo formacii
			if (GetFormationPos(defpos) >= 0) minidx = 1;

			if (minidx >= 0) {
				m_pPerception.mm_x2 = (float)defpos.GetX();
				m_pPerception.mm_y2 = (float)defpos.GetY();

				if ((pThisPlayer->GetPosition() - defpos).GetLength() > 3) {
					if (RunTo(defpos, 1) > 0) minidx = -1;
					contrun = defpos;  cont = 8;
				}
			} 

			if (minidx < 0)
				Turn(pThisPlayer->GetAngleTo(m_pPerception.GetBall()));
		}
	}
}

/*
void CBaseAgent::AutoPlay1()
{
	static int letsee = 0;
	static int cnt = 0;
	TPlayer* pThisPlayer = m_pPerception.GetThisPlayer();

//m_aPlayer.Turn(30);
//return;

	if (!m_pPerception.GetSeeBall()) {
		if (letsee) Turn(pThisPlayer->GetAngleTo(m_pPerception.GetBall()));
		else Turn(-45);
		letsee = 0;		
	} else {
		letsee = 1;
		double dist = pThisPlayer->GetDistanceTo(m_pPerception.GetBall());
		TVector ballspeed = m_pPerception.GetBall()->GetAbsSpeed();
		TVector newpos = GetBallInterceptPos();
		if ((dist < 1.0) || (RunTo(newpos))) {
			// toc loptu, len ked stojime
//			if (pThisPlayer->GetAbsSpeed().GetLength() < 0.01) 
			if (cnt++ > 8) { Turn(90); cnt = 0; }
			else
			  TurnBallTo(m_pPerception.GetGoalDirection());
//			KickTo(m_pPerception.GetGoalDirection());
		} else cnt = 0;
	}
}
*/

// Pili
void CBaseAgent::AutoPlay2()
{
	double ToPointX,ToPointY,BallX,BallY,GoalX,GoalY,GoalToBallDist,
		   GoalToBallNormDistX,GoalToBallNormDistY,
		   PlayerX,PlayerY,Dist;
	double RequestedDistFromBall = 1.5;
	double Tolerance=0.3;
	int ToDir;

	if (!m_pPerception.GetBall()->IsValid()) {
		Turn(15);
		return;
	}

	BallX = m_pPerception.GetBall()->GetAbsX();
	BallY = m_pPerception.GetBall()->GetAbsY();

	PlayerX = m_pPerception.GetThisPlayer()->GetAbsX();
	PlayerY = m_pPerception.GetThisPlayer()->GetAbsY();
	
	GoalX = 105/2;
	GoalY = 0;

	GoalToBallDist = sqrt( (GoalX-BallX)*(GoalX-BallX) + (GoalY-BallY)*(GoalY-BallY) );
	GoalToBallNormDistX = (GoalX - BallX) / GoalToBallDist;
	GoalToBallNormDistY = (GoalY - BallY) / GoalToBallDist;
	ToPointX = BallX - GoalToBallNormDistX*RequestedDistFromBall;
	ToPointY = BallY - GoalToBallNormDistY*RequestedDistFromBall;

	Dist = sqrt( (ToPointX-PlayerX)*(ToPointX-PlayerX) + (ToPointY-PlayerY)*(ToPointY-PlayerY) );
	if (Dist <= Tolerance) {
		ToDir = m_pPerception.GetGoalDirection();
		Kick(50,ToDir);
	} else {
		ToDir = m_pPerception.GetThisPlayer()->GetAngleTo(&TVisibleObject(ToPointX,ToPointY));
		if ((ToDir > 2) || (ToDir<-2)) Turn(ToDir);
		else {
			if (Dist>1.25*Tolerance) Dash(10);
			else Dash(5);
		};
	}
}

BOOL CBaseAgent::IsBallCaught()
{
	int Side,Mode;

	Side = m_pPerception.GetSide();
	Mode = m_pPerception.GetPlayMode(); 
	if ( (Side == LEFT_TO_RIGHT) && (Mode == M_FREE_KICK_L)) return TRUE;
	if ( (Side == RIGHT_TO_LEFT) && (Mode == M_FREE_KICK_R)) return TRUE;
	return FALSE;
}

BOOL CBaseAgent::CanCatchBall(double dPlayerX,double dPlayerY,double dBallX,double dBallY)
{
	double TmpX,TmpY;

	TmpX = fabs(dBallX-dPlayerX);
	TmpY = fabs(dBallY-dPlayerY);

	if ( (TmpX <= CATCHABLE_AREA_L/2.0) && (TmpY <= CATCHABLE_AREA_W/2.0) ) return TRUE;
	return FALSE;
}

// Goalkeeper
void CBaseAgent::AutoPlay3()
{
	TVisibleObject *pBall;
	TPlayer *pThisPlayer;
	static BOOL TurnToLastBallPos = FALSE;
	static BOOL InitPosOK = FALSE;
	static BOOL GoToHome = FALSE;
	static int KickNum = 0;
	static TVector *pKeeperHome=NULL;
	static int KeeperSide = -1;
	const double GroundSizeX = 150;
	const double GoalArea = 15;
	const double HomePointTolerance = 2.5;
	const double KickDist = 0.4;
	const int FindBallTurn = 30;
	const int AngleTolerance = 5;

	pThisPlayer = m_pPerception.GetThisPlayer();
	pBall = m_pPerception.GetBall();
	TVector NewBallPos = GetBallInterceptPos(pThisPlayer);

	// initialize keeper home
	if (pKeeperHome != NULL) {	
		if (m_pPerception.GetSide() != KeeperSide) {
			delete pKeeperHome;
			pKeeperHome = NULL;
		}
	}
	if (pKeeperHome == NULL) {
		KeeperSide = m_pPerception.GetSide();
		if (KeeperSide == LEFT_TO_RIGHT) pKeeperHome = new TVector(-105.0/2+5,0.0);
		else pKeeperHome = new TVector(105.0/2-5,0.0);
	}

	// if mode is M_BEFORE_KICK_OFF, then move keeper to home
	if ( (m_pPerception.GetPlayMode() == M_BEFORE_KICK_OFF) && (!InitPosOK) ) {
		Move((int)pKeeperHome->GetX(),(int)pKeeperHome->GetY());
		InitPosOK = TRUE;
		return;
	}
	if (m_pPerception.GetPlayMode() == M_BEFORE_KICK_OFF) return;

	// if we caught ball
	if (IsBallCaught()) {
		Kick(100,m_pPerception.GetGoalDirection());
		return;
	}

	if (GoToHome)
		if (pThisPlayer->GetDistanceTo(*pKeeperHome) > HomePointTolerance) {
			int KeeperHomeDir = pThisPlayer->GetAngleTo(*pKeeperHome);
			if (abs(KeeperHomeDir) > AngleTolerance) Turn(KeeperHomeDir);
			else Dash(100);
			return;
		} else GoToHome = FALSE;

	// if ball isn't visible
	if (!m_pPerception.GetSeeBall()) {
		if (TurnToLastBallPos) Turn(pThisPlayer->GetAngleTo(pBall->GetLastSeePosition()));
		else Turn(FindBallTurn);
		TurnToLastBallPos = FALSE;
		return;
	} else TurnToLastBallPos = TRUE;

	// if we can catch the ball
	if (CanCatchBall(pThisPlayer->GetAbsX(),pThisPlayer->GetAbsY(),pBall->GetAbsX(),pBall->GetAbsY())) {
		Catch(pThisPlayer->GetAngleTo(NewBallPos));
		return;
	} // if

	// ball is valid, but can't be caught, so turn directly to it (if we are not)
	if (abs(pThisPlayer->GetAngleTo(NewBallPos)) > AngleTolerance) {
		Turn(pThisPlayer->GetAngleTo(NewBallPos));
		return;
	} // if

	// we can see the ball but we can't catch it

	// if the ball is in our watch area, we must run there
    if (pBall->GetDistanceTo(*pKeeperHome) <= GoalArea) {
		Dash(100);
		return;
	} else GoToHome = TRUE;
}

int CompareInt(const void* elem1, const void* elem2)
{
	return *((int*) elem1) - *((int*) elem2);
}

// Rado + Dusky
void CBaseAgent::AutoPlay4()
{
	static int letsee = 0;

//m_aPlayer.Turn(30);
//return;

	if (!m_pPerception.GetSeeBall()) {
		if (letsee) Turn(m_pPerception.GetThisPlayer()->GetAngleTo(m_pPerception.GetBall()));
		else Turn(-45);
		letsee = 0;		
	} else {
		letsee = 1;
		int dir = m_pPerception.GetThisPlayer()->GetAngleTo(m_pPerception.GetBall());
//		int dir = m_pPerception.m_pBall->GetDirection();
		if ((dir > 2) || (dir < -2)) Turn(dir);
		else {
			double dist = m_pPerception.GetBall()->GetDistance();
			if (dist > 1.6) Dash(100);
			else {
				int dir2 = m_pPerception.GetGoalDirection();
				if (dir2 >  90) Kick(20,  90); else 
				if (dir2 < -90) Kick(20, -90); else
//-------------------------------------
				{

/*				char buffer[100];
				sprintf(buffer, "%lf", m_pPerception.GetGoalDistance());
				Say(buffer);
*/

				if (m_pPerception.GetGoalDistance() > 30)
					Kick(100, dir2);
				else
				{
					int	 nIndex;

					if (m_pPerception.GetSide() == LEFT_TO_RIGHT)
						nIndex = GOAL_POST_R;
					else
						nIndex = GOAL_POST_L;
					
					int aAngle[MAX_PLAYERS + 2];

					aAngle[0] = m_pPerception.GetThisPlayer()->GetAngleTo(m_pPerception.GetFlags()[nIndex]);
					aAngle[1] = m_pPerception.GetThisPlayer()->GetAngleTo(m_pPerception.GetFlags()[nIndex + 1]);

					if (aAngle[0] > aAngle[1])
					{
						int nTmp = aAngle[0];
						aAngle[0] = aAngle[1];
						aAngle[1] = nTmp;
					}

					int nCount = 2;
					int i;

					for (i = 0; i < m_pPerception.GetNumOfVisiblePlayers(); i++)
					{
						int nAngle = m_pPerception.GetThisPlayer()->GetAngleTo(m_pPerception.GetVisiblePlayers()[i]);
						if ((nAngle > aAngle[0]) && (nAngle < aAngle[1]))
							aAngle[nCount++] = nAngle;
					}

					qsort(aAngle, nCount, sizeof (int), ::CompareInt);

					int nMaxGap = 0;
					int nGapNdx;

					for (i = 0; i < (nCount - 1); i++)
					{
						int nGap = aAngle[i + 1] - aAngle[i];
						if (nGap > nMaxGap)
						{
							nMaxGap = nGap;
							nGapNdx = i;
						}
					}

//					int goaldir = m_pPerception.GetGoalDirection();
					Kick(100, (aAngle[nGapNdx + 1] + aAngle[nGapNdx]) / 2);
				}
				}
//-------------------------------------
			}
		}		
	}
}

void CBaseAgent::AutoPlay()
{
	switch (GetAutoPlay()) {
		case 0:
			break;
		case 1:
			AutoPlay1();
			break;
		case 2:
			AutoPlay2();
			break;
		case 3:
			AutoPlay3();
			break;
		case 4:
			AutoPlay4();
			break;
	}
}

int CBaseAgent::SetAutoPlay(int aNewMode)
{
	int OldMode;

	OldMode=GetAutoPlay();
	m_iAutoPlay = aNewMode;
	return OldMode;
}

int CBaseAgent::AbsTurn(int aNewAbsDir)
{
	int OldAbsDir = m_pPerception.GetThisPlayer()->GetAbsDir();

	Turn(aNewAbsDir-OldAbsDir);
	return OldAbsDir;
}